#include "miniaudioengine.h"
#include "dmxPersonality.h"

#define BIAS                0.99f
#define FILTER_ORDER        3

static ma_pcm_rb aux1Buffer;
static ma_data_source_node g_dataSupplyNode;
static ma_data_source_rb g_dataSourceRB;

MiniAudioEngine::MiniAudioEngine() {}

void MiniAudioEngine::audioDataCallback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    ma_result result;

    result = ma_engine_read_pcm_frames((ma_engine*)pDevice->pUserData, pOutput, frameCount, NULL);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": error audio callback.";
    }
    (void)pInput;
}

void MiniAudioEngine::audioDataCallback1(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    (void)pDevice;
    ma_result result;

    ma_uint32 pcmFramesAvailableInRB = 0;
    ma_uint32 pcmFramesProcessed = 0;
    while (pcmFramesProcessed < frameCount) {
        pcmFramesAvailableInRB = ma_pcm_rb_available_read(&aux1Buffer);
        if (pcmFramesAvailableInRB == 0) {
            break;
        }
        ma_uint32 framesToRead = frameCount - pcmFramesProcessed;
        if (framesToRead > pcmFramesAvailableInRB) {
            framesToRead = pcmFramesAvailableInRB;
        }
        void* pReadBuffer = NULL;
        ma_pcm_rb_acquire_read(&aux1Buffer, &framesToRead, &pReadBuffer);
        if (pReadBuffer != NULL) {
            ma_copy_pcm_frames(pOutput, pReadBuffer, framesToRead, FORMAT, CHANNELS);
            ma_pcm_rb_commit_read(&aux1Buffer, framesToRead);
            pcmFramesProcessed += framesToRead;
        }/* else { break; }*/
    }
    (void)pInput;
}

void MiniAudioEngine::stopEngine()
{
    for (uint i = 0; i < m_layersQty; i++) {
        ma_sound_uninit(&m_currentSound[i]);
    }
    for (uint i = 0; i < m_devicesSelected; i++) {
        for (uint j = 0; j < m_layersQty; j++) {
            ma_node_uninit(&m_filterBank[i][j].input, NULL);
            ma_hpf_node_uninit(&m_filterBank[i][j].hpf, NULL);
            ma_loshelf_node_uninit(&m_filterBank[i][j].loshelf, NULL);
            ma_peak_node_uninit(&m_filterBank[i][j].mLow, NULL);
            ma_peak_node_uninit(&m_filterBank[i][j].mHigh, NULL);
            ma_hishelf_node_uninit(&m_filterBank[i][j].hishelf, NULL);
            ma_splitter_node_uninit(&m_filterBank[i][j].output, NULL);
        }
        //ma_writer_node_uninit(&m_sendToAux[0], NULL);
        //ma_pcm_rb_uninit(&aux1Buffer);
        ma_engine_uninit(&m_engine[i]);
        ma_device_uninit(&m_device[i]);
    }
    ma_context_uninit(&m_context);
    ma_resource_manager_uninit(&m_resourceManager);
}

bool MiniAudioEngine::startEngine(uint layers)
{
    ma_result result;

    m_layersQty = layers;
    for (uint i =0; i < m_layersQty; i++) {
        m_mediaLoaded[i] = false;
        m_currentLayerValues[i].status = Status::Iddle;
        m_currentLayerValues[i].pan = 128;
        m_currentLayerValues[i].pitch = 128;
        m_currentLayerValues[i].vol = 0;
        m_currentLayerValues[i].cursor = 0;
    }
    result = this->startContext();
    if (result != MA_SUCCESS) return false;
    result = this->getAllAudioDevices();
    if (result != MA_SUCCESS) return false;
    return true;
}

ma_result MiniAudioEngine::createFilterBank(int id, uint layer)
{
    ma_result result;
    ma_node_graph *ng = ma_engine_get_node_graph(&m_engine[id]);
    ma_node *endpoint = ma_engine_get_endpoint(&m_engine[id]);
    filterBank *fb    = &m_filterBank[id][layer];

    ma_splitter_node_config splitterConfig = ma_splitter_node_config_init(CHANNELS);
    result = ma_splitter_node_init(ng, &splitterConfig, NULL, &fb->input);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init input node." << endl;
        return result;
    }

    fb->hpfConfig = ma_hpf_node_config_init(CHANNELS, SAMPLE_RATE, 16, FILTER_ORDER);
    result = ma_hpf_node_init(ng, &fb->hpfConfig, NULL, &fb->hpf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init high pass filter node." << endl;
        return result;
    }

    fb->loshelfConfig = ma_loshelf_node_config_init(CHANNELS, SAMPLE_RATE, 0.0f, 1.0f, 30);
    result = ma_loshelf_node_init(ng, &fb->loshelfConfig, NULL, &fb->loshelf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init low pass filter node." << endl;
        return result;
    }

    fb->mLowConfig = ma_peak_node_config_init(CHANNELS, SAMPLE_RATE, 0.0, 4.0, 200); // double gainDB, double q, double frequency);
    result = ma_peak_node_init(ng, &fb->mLowConfig, NULL, &fb->mLow);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init peak low filter node." << endl;
        return result;
    }

    fb->mHighConfig = ma_peak_node_config_init(CHANNELS, SAMPLE_RATE, 0.0, 0.0, 600); // double gainDB, double q, double frequency);
    result = ma_peak_node_init(ng, &fb->mHighConfig, NULL, &fb->mHigh);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init peak high filter node." << endl;
        return result;
    }

    fb->hishelfConfig = ma_hishelf_node_config_init(CHANNELS, SAMPLE_RATE, 0.0f, 1.0f, 20000);
    result = ma_hishelf_node_init(ng, &fb->hishelfConfig, NULL, &fb->hishelf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init hi shelf filter node." << endl;
        return result;
    }

    result = ma_splitter_node_init(ng, &splitterConfig, NULL, &fb->output);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to init output node." << endl;
        return result;
    }

    result = ma_node_attach_output_bus(&fb->input, 0, &fb->hpf, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach input node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->input, 1, &fb->output, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach bypass connection." << endl;
        return result;
    }
    ma_node_set_output_bus_volume(&fb->input, 1, 0.0f);
    result = ma_node_attach_output_bus(&fb->hpf, 0, &fb->loshelf, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach high pass pass filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->loshelf, 0, &fb->mLow, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach low shelf filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->mLow, 0, &fb->mHigh, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach low peaks filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->mHigh, 0, &fb->hishelf, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach high peaks filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&fb->hishelf, 0, &fb->output, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach high shelf filter node." << endl;
        return result;
    }
if (id == 0) {
        result = ma_node_attach_output_bus(&fb->output, 0, &m_sendToAux[id], 0);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to attach output node to engine." << endl;
            return result;
        }
        result = ma_node_attach_output_bus(&fb->output, 1, &m_sendToAux[id], 1);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to attach output node to aux send 1." << endl;
            return result;
        }
    }
    return result;
}

ma_result MiniAudioEngine::setNodeGraph(int id) {
    ma_result result = MA_SUCCESS;
    uint i = 0;

    if (id == 0) {
        ma_node_graph *ng = ma_engine_get_node_graph(&m_engine[id]);
        size_t sizeInFrames = SAMPLE_RATE; // ma_get_bytes_per_frame(FORMAT, CHANNELS);
        result = ma_pcm_rb_init(FORMAT, CHANNELS, sizeInFrames, NULL, NULL, &aux1Buffer);
        if (result != MA_SUCCESS) {
            printf("Failed to initialize ring buffer.\n");
            return result;
        }
        ma_silence_pcm_frames(aux1Buffer.rb.pBuffer, sizeInFrames, FORMAT, CHANNELS);
        ma_writer_node_config writerConfig = ma_writer_node_config_init(CHANNELS, SAMPLE_RATE * 5, &aux1Buffer);
        result = ma_writer_node_init(ng, &writerConfig, NULL, &m_sendToAux[id]);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to init writer node." << endl;
            return result;
        }
        result = ma_node_attach_output_bus(&m_sendToAux[id], 0, ma_engine_get_endpoint(&m_engine[id]), 0); // Pull API
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to attach writer node." << endl;
            return result;
        }
    }
    while (result == MA_SUCCESS && i < m_layersQty) {
        result = this->createFilterBank(id, i);
        i++;
    }
    return (result);
}

bool MiniAudioEngine::startDevice(uint *systemId, uint nb)
{
    ma_result result = MA_SUCCESS;
    ma_device_config deviceConfig;
    ma_engine_config engineConfig;

    m_devicesSelected = nb;
    for (uint internalId = 0; internalId < nb; internalId++) {
        deviceConfig = ma_device_config_init(ma_device_type_duplex);
        deviceConfig.capture.pDeviceID  = &m_pPlaybackDeviceInfos[systemId[internalId]].id;
        deviceConfig.capture.format     = m_resourceManager.config.decodedFormat;
        deviceConfig.capture.channels   = CHANNELS;
        deviceConfig.capture.shareMode  = ma_share_mode_shared;
        deviceConfig.playback.pDeviceID = &m_pPlaybackDeviceInfos[systemId[internalId]].id;
        deviceConfig.playback.format    = m_resourceManager.config.decodedFormat;
        deviceConfig.playback.channels  = CHANNELS;
        deviceConfig.sampleRate         = m_resourceManager.config.decodedSampleRate;
        if (internalId == 0)
            deviceConfig.dataCallback   = audioDataCallback;
        else if (internalId == 1)
            deviceConfig.dataCallback   = audioDataCallback1;
        deviceConfig.pUserData          = &m_engine[internalId];
        result = ma_device_init(&m_context, &deviceConfig, &m_device[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to initialize audio device " << m_pPlaybackDeviceInfos[*systemId].name << endl;
            return false;
        }
        engineConfig                  = ma_engine_config_init();
        engineConfig.pDevice          = &m_device[internalId];
        engineConfig.pResourceManager = &m_resourceManager;
        engineConfig.gainSmoothTimeInMilliseconds = SAMPLE_RATE / 100;
        engineConfig.noAutoStart      = MA_TRUE;
        result = ma_engine_init(&engineConfig, &m_engine[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to initialize audio engine" << endl;
            return false;
        }
        result = this->setNodeGraph(internalId);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to set node graph " << systemId[internalId] << endl;
            return false;
        }
        result = ma_engine_start(&m_engine[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to start audio engine" << systemId[internalId] << endl;
            return false;
        }
        cout << "Initialized Audio Device. internalId: " << internalId << " systemId: " << systemId[internalId] << " " << m_pPlaybackDeviceInfos[systemId[internalId]].name << endl;
    }
    //result = ma_data_source_rb_init(&g_dataSourceRB, &aux1Buffer);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to init data source ring buffer" << endl;
        return false;
    }
    ma_data_source_node_config dataSupplyNodeConfig = ma_data_source_node_config_init(&g_dataSourceRB);
    //result = ma_data_source_node_init(ma_engine_get_node_graph(&m_engine[1]), &dataSupplyNodeConfig, NULL, &g_dataSupplyNode);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to init data source node" << endl;
        return false;
    }
    //result = ma_node_attach_output_bus(&g_dataSupplyNode, 0, ma_engine_get_endpoint(&m_engine[1]), 0);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to attach data source rb node" << endl;
        return false;
    }
    cout << "data source node state " << ma_node_get_state(&g_dataSupplyNode.base) << endl;
    return true;
}

ma_result MiniAudioEngine::startContext()
{
    ma_result result;

    m_resourceManagerConfig = ma_resource_manager_config_init();
    m_resourceManagerConfig.decodedFormat     = FORMAT;
    m_resourceManagerConfig.decodedChannels   = CHANNELS;
    m_resourceManagerConfig.decodedSampleRate = SAMPLE_RATE;
    m_resourceManagerConfig.jobThreadCount    = 4;
    result = ma_resource_manager_init(&m_resourceManagerConfig, &m_resourceManager);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to initialize audio resource manager." << endl;
        return result;
    }
    result = ma_context_init(NULL, 0, NULL, &m_context);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to initialize audio context." << endl;
    }
    return result;
}

// enum all audio devices in system
ma_result MiniAudioEngine::getAllAudioDevices()
{
    ma_result result;

    result = ma_context_get_devices(&m_context, &m_pPlaybackDeviceInfos, &m_playbackDeviceCount, NULL, NULL);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to enumerate playback devices." << endl;
        ma_context_uninit(&m_context);
        return result;
    }
    cout << "Audio devices available:" << endl;
    for (ma_uint32 iAvailableDevice = 0; iAvailableDevice < m_playbackDeviceCount; iAvailableDevice += 1) {
        cout << iAvailableDevice << " : " << m_pPlaybackDeviceInfos[iAvailableDevice].name << endl;
    }
    return result;
}

ma_result MiniAudioEngine::loadMedia(int layer, char *file, uint audioDevice)
{
    ma_result result;

    // - iniciar un sonido por cada capa, copiar la capa en otro dispositivo
    // - writer node y source con el buffer escrito en el otro device
    if (m_mediaLoaded[layer] == true)
    {
        ma_sound_uninit(&m_currentSound[layer]);
        m_mediaLoaded[layer] = false;
    }
    result = ma_sound_init_from_file(&m_engine[audioDevice], file, \
                MA_SOUND_FLAG_NO_SPATIALIZATION \
                , NULL, NULL, &m_currentSound[layer]);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to load file " << file << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&m_currentSound[layer], 0, &m_filterBank[audioDevice][layer].input, 0);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to attach sound output bus " << audioDevice << endl;
        //return result;
    }
    m_mediaLoaded[layer] = true;
    this->refreshValues(layer);
    m_currentLayerValues[layer].media = file;
    return result;
}

float MiniAudioEngine::getDuration(int layer)
{
    ma_result result;
    float ret;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS) {
        return result;
    }
    return (ret * 1000);
}

float MiniAudioEngine::getCursor(int layer)
{
    ma_result result;
    float ret = 0;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_cursor_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS)
    {
        cout << "Error" << result << ": Can not get cursor " << layer << endl;
        ret = MA_ERROR;
    }
    return ret;
}

ma_result MiniAudioEngine::printFormatInfo(int layer)
{
    ma_format format;
    ma_uint32 channels;
    ma_uint32 sampleRate;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    ma_result result = ma_sound_get_data_format(&m_currentSound[layer], \
                                    &format, &channels, &sampleRate, NULL, 0);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to get data format " << layer;
        cout << endl;
    } else {
        cout << "Layer:" << layer << " ";
        cout << m_currentLayerValues[layer].media.toLatin1().data();
        cout << " samples/sec:" << sampleRate << " format:" <<  format;
        cout << " channels:" << channels << endl;
    }
    return result;
}

// Expects between 0 and 1 vol value
void MiniAudioEngine::volChanged(int layer, float vol)
{
    if (m_mediaLoaded[layer] == false)
        return;
    if (vol >= 1)
        vol = 0.99f;
    ma_sound_group_set_fade_in_milliseconds(&m_currentSound[layer], -1, pow(vol, 3), FADE_TIME);
    m_currentLayerValues[layer].vol = vol;
}

void MiniAudioEngine::panChanged(int layer, float value)
{
    float result;

    if (m_mediaLoaded[layer] == false)
        return;
    result = (value / 128.0) - 1.0;
    ma_sound_group_set_pan(&m_currentSound[layer], result);
    m_currentLayerValues[layer].pan = value;
}

void MiniAudioEngine::pitchChanged(int layer, float value)
{
    float pitch;

    if (m_mediaLoaded[layer] == false)
        return;
    pitch = value / 128.0;
    ma_sound_group_set_pitch(&m_currentSound[layer], pitch);
    m_currentLayerValues[layer].pitch = value;
}

ma_result MiniAudioEngine::playbackChanged(int layer, Status status)
{
    ma_result result = MA_SUCCESS;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    bool loop = false;
    switch (status) {
        case Status::Paused:
            result = ma_sound_stop_with_fade_in_milliseconds(&m_currentSound[layer], FADE_TIME);
            break;
        case Status::Stopped:
            ma_sound_stop_with_fade_in_milliseconds(&m_currentSound[layer], FADE_TIME);
            result = this->seekToCursor(layer, m_currentLayerValues[layer].cursor);
            break;
        case Status::PlayingLoop:
            loop = true;
        case Status::PlayingOnce:
        case Status::PlayingFolder:
        case Status::PlayingFolderLoop:
        case Status::PlayingFolderRandom:
            ma_sound_set_stop_time_in_milliseconds(&m_currentSound[layer], ~(ma_uint64)0);
            ma_sound_set_looping(&m_currentSound[layer], loop);
            result = ma_sound_start(&m_currentSound[layer]);
            if (m_currentLayerValues[layer].cursor != 0) {
                usleep(1000 * 50); // Avoid small glitch at start, how to flush the cached buffers in audio pipe line?
            }
            this->volChanged(layer, m_currentLayerValues[layer].vol);
        default:
            break;
    }
    if (result == MA_SUCCESS)
        m_currentLayerValues[layer].status = status;
    return result;
}

ma_result MiniAudioEngine::seekToCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;
    ma_uint64 end, start;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_pcm_frames(&m_currentSound[layer], &end);
    if (result != MA_SUCCESS) { return result; }
    start = (cursor * end) / 65025;
    result = ma_sound_seek_to_pcm_frame(&m_currentSound[layer], start);
    //result = ma_data_source_set_loop_point_in_pcm_frames(&m_currentSound[layer], start, end); // this do nothing here, it must be done after set_looping or start?
    return (result);
}

ma_result MiniAudioEngine::setCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;

    m_currentLayerValues[layer].cursor = cursor;
    result = this->seekToCursor(layer, cursor);
    return (result);
}

Status MiniAudioEngine::getStatus(int layer)
{
    return m_currentLayerValues[layer].status;
}

void MiniAudioEngine::refreshValues(int layer)
{
    this->seekToCursor(layer, m_currentLayerValues[layer].cursor);
    this->panChanged(layer, m_currentLayerValues[layer].pan);
    this->volChanged(layer, m_currentLayerValues[layer].vol);
    this->pitchChanged(layer, m_currentLayerValues[layer].pitch);
    this->playbackChanged(layer, m_currentLayerValues[layer].status);
}

ma_result MiniAudioEngine::filterParamChanged(int layer, int audioDevice, int channel, int value)
{
    ma_result result = MA_SUCCESS;

    filterBank *fb = &m_filterBank[audioDevice][layer];

    if (channel == HP_FREQ) {
        fb->hpfConfig.hpf.cutoffFrequency = double((value * 1.31) + 16.0f); // 16 - 350
        result = ma_hpf_node_reinit(&fb->hpfConfig.hpf, &fb->hpf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency high pass filter node." << endl;
            return result;
        }
    } else if (channel == LOW_FREQ) {
        fb->loshelfConfig.loshelf.frequency = 30 + (value * 1.647); // 30 - 450
        result = ma_loshelf_node_reinit(&fb->loshelfConfig.loshelf, &fb->loshelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency low shelf filter node." << endl;
            return result;
        }
    } else if (channel == LOW_Q) {
        fb->loshelfConfig.loshelf.shelfSlope = (double)(value / 32.0f) + 0.1f; // 0.1 - 8
        result = ma_loshelf_node_reinit(&fb->loshelfConfig.loshelf, &fb->loshelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set Q low shelf filter node." << endl;
            return result;
        }
    } else if (channel == LOW_GAIN) {
        fb->loshelfConfig.loshelf.gainDB = (double)(value / 21.25f) - 6.023528412f;
        result = ma_loshelf_node_reinit(&fb->loshelfConfig.loshelf, &fb->loshelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set gain low shelf filter node." << endl;
            return result;
        }
    } else if (channel == MIDLOW_FREQ) {
        fb->mLowConfig.peak.frequency = 200 + (value * 9.019607843); // 200 - 450
        result = ma_peak_node_reinit(&fb->mLowConfig.peak, &fb->mLow);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency Mid Low pass filter node." << endl;
            return result;
        }
    } else if (channel == MIDLOW_Q) {
        fb->mLowConfig.peak.q = (double)( value / 64.0f) + 0.10; // 0.1 - 4
        result = ma_peak_node_reinit(&fb->mLowConfig.peak, &fb->mLow);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set Q Mid Low filter node." << endl;
            return result;
        }
    } else if (channel == MIDLOW_GAIN) {
        fb->mLowConfig.peak.gainDB = (double)(value / 7.0833333333333f) - 18.0f;
        result = ma_peak_node_reinit(&fb->mLowConfig.peak, &fb->mLow);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set gain Mid Low filter node." << endl;
            return result;
        }
    } else if (channel == MIDHIGH_FREQ) {
        fb->mHighConfig.peak.frequency = 600 + (value * 25.09803922); // 600 - 7000
        result = ma_peak_node_reinit(&fb->mHighConfig.peak, &fb->mHigh);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set frecuency Mid High filter node." << endl;
            return result;
        }
    } else if (channel == MIDHIGH_Q) {
        fb->mHighConfig.peak.q = (double)( value / 64.0f) + 0.10; // 0.1 - 4
        result = ma_peak_node_reinit(&fb->mHighConfig.peak, &fb->mHigh);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set Q Mid High filter node." << endl;
            return result;
        }
    } else if (channel == MIDHIGH_GAIN) {
        fb->mHighConfig.peak.gainDB = (double)(value / 7.0833333333333f) - 18.0f;
        result = ma_peak_node_reinit(&fb->mHighConfig.peak, &fb->mHigh);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed to set gain Mid High filter node." << endl;
            return result;
        }
    }  else if (channel == HIGH_FREQ) {
    fb->hishelfConfig.hishelf.frequency = 1500 + (value * 56.8627451); // 1500 - 16000
    result = ma_hishelf_node_reinit(&fb->hishelfConfig.hishelf, &fb->hishelf);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to frecuency high shelf filter node." << endl;
        return result;
    }
    } else if (channel == HIGH_Q) {
        fb->hishelfConfig.hishelf.shelfSlope = (double)( value / 32.0f) + 0.1f;
        result = ma_hishelf_node_reinit(&fb->hishelfConfig.hishelf, &fb->hishelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set Q high shelf filter node." << endl;
            return result;
        }
    } else if (channel == HIGH_GAIN) {
        fb->hishelfConfig.hishelf.gainDB = (double)(value / 21.25) - 6.023528412f;
        result = ma_hishelf_node_reinit(&fb->hishelfConfig.hishelf, &fb->hishelf);
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set gain high shelf filter node." << endl;
            return result;
        }
    } else if (channel == SEND1) {
        ma_node_set_output_bus_volume(&fb->output, 0, pow((value / 255.0f), 2));
        if (result != MA_SUCCESS) {
            cout << "ERROR " << result << ": Failed set Send 1 Volume." << endl;
            return result;
        }
    }  else if (channel == SEND2) {
        ma_node_set_output_bus_volume(&fb->output, 1, pow((value / 255.0f), 2));
       if (result != MA_SUCCESS) {
           cout << "ERROR " << result << ": Failed set Send 2 Volume." << endl;
       }
       return result;
   }
   return (result);
}

bool MiniAudioEngine::setBypass(int audioDevice, int layer, bool bypass)
{
    filterBank *fb = &m_filterBank[audioDevice][layer];

    if (bypass) {
        ma_node_set_output_bus_volume(&fb->input, 1, 1.0f);
        ma_node_set_output_bus_volume(&fb->input, 0, 0.0f);
    } else {
        ma_node_set_output_bus_volume(&fb->input, 1, 0.0f);
        ma_node_set_output_bus_volume(&fb->input, 0, 1.0f);
    }

}
