#ifndef MINIAUDIOENGINE_H
#define MINIAUDIOENGINE_H

#define MA_ENABLE_ONLY_SPECIFIC_BACKENDS
#define MA_ENABLE_JACK
#define MA_NO_GENERATION
#define MA_DEBUG_OUTPUT
#define MA_DISABLE_PULSEAUDIO
#define MA_DEBUG_OUTPUT
#define MA_LOG_LEVEL_DEBUG DEBUG
#define MINIAUDIO_IMPLEMENTATION
#include "miniaudio.h"
#include "ma_writer_node.h"
#include <bits/stdc++.h>
using namespace std;
#include "defines.h"


typedef struct
{
    ma_splitter_node input;
    ma_hpf_node hpf;
    ma_hpf_node_config hpfConfig;
    ma_loshelf_node loshelf;
    ma_loshelf_node_config loshelfConfig;
    ma_peak_node mLow;
    ma_peak_node_config mLowConfig;
    ma_peak_node mHigh;
    ma_peak_node_config mHighConfig;
    ma_hishelf_node hishelf;
    ma_hishelf_node_config hishelfConfig;
    ma_splitter_node output;
    ma_writer_node send1;
    ma_data_source_node return1;
    ma_audio_buffer_ref supplyReturn1;
} filterBank;

typedef struct
{
   ma_engine m_engine[MAX_AUDIODEVICES];
   filterBank m_filterBank[MAX_AUDIODEVICES][MAX_LAYERS];
   ma_writer_node m_sendAux1[MAX_LAYERS];
   ma_pcm_rb *aux1Buffer;
} audioObjects;

class MiniAudioEngine
{
    friend class libreMediaServerAudio;
    static ma_pcm_rb *rb;

public:
    MiniAudioEngine();
    void stopEngine();
    bool startEngine(uint layersQty);
    bool startDevice(uint *id, uint nb);
    static void audioDataCallback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount);
    static void audioDataCallback1(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount);
    bool setBypass(int audioDevice, int layer, bool bypass);

protected:
    ma_result loadMedia(int layer, char *media, uint audioDevice);
    void volChanged(int layer, float vol);
    void panChanged(int layer, float pan);
    void pitchChanged(int layer, float pitch);
    ma_result playbackChanged(int layer, Status status);
    ma_result setCursor(int layer, int cursor);
    ma_result printFormatInfo(int layer);
    float getDuration(int layer);
    float getCursor(int layer);
    Status getStatus(int layer);
    inline float getVol(int layer) {
        return ma_sound_get_volume(&m_currentSound[layer]); }
    inline bool getAtEnd(int layer) { return m_currentSound[layer].atEnd; }
    ma_result filterParamChanged(int layer, int audioDevice, int channel, int value);

private:
    ma_resource_manager_config m_resourceManagerConfig;
    ma_resource_manager m_resourceManager;
    ma_context m_context;
    ma_device_info* m_pPlaybackDeviceInfos;
    ma_uint32 m_playbackDeviceCount;
    ma_uint32 m_devicesSelected;
    ma_device m_device[MAX_AUDIODEVICES];
    ma_sound m_currentSound[MAX_LAYERS];
    ma_bool8 m_mediaLoaded[MAX_LAYERS];
    layerData m_currentLayerValues[MAX_LAYERS];
    filterBank m_filterBank[MAX_AUDIODEVICES][MAX_LAYERS];
    ma_engine m_engine[MAX_AUDIODEVICES];
    uint m_layersQty;
    ma_writer_node m_sendToAux[MAX_LAYERS];
    audioObjects m_audioObjects;

    ma_result getAllAudioDevices();
    ma_result startContext();
    void refreshValues(int layer);
    ma_result seekToCursor(int layer, int cursor);
    ma_result setNodeGraph(int id);
    ma_result createFilterBank(int id, uint layer);
};

#endif // MINIAUDIOENGINE_H
